#!/bin/bash

units=(ui-utils plugin-model plugin-ui utils)

for fl in ${units[@]} ; do
    compute-theme-d-pcode-deps $fl.thi "intr-gtk4-examples calc2"
    compute-theme-d-pcode-deps $fl.thb "intr-gtk4-examples calc2"
done

compute-theme-d-pcode-deps main.thp "intr-gtk4-examples calc2"

for fl in ${units[@]} ; do
    compute-theme-d-module-deps $fl.thi "intr-gtk4-examples calc2"
    compute-theme-d-module-deps $fl.thb "intr-gtk4-examples calc2"
done

compute-theme-d-program-deps --module main.thp "intr-gtk4-examples calc2"
