(intr-entities
 (only-types
  (GLib String))
 (classes
  (Gio BusType)
  (Gio BusNameOwnerFlags)
  (Gio DBusMethodInfo)
  (Gio DBusPropertyInfo)
  (Gio DBusSignalInfo)
  (Gio DBusInterfaceInfo))
 (functions
  (Gio bus_own_name))
 (rejected-methods ref unref generate-xml))
