
;; Copyright (C) 2026 Tommi Höynälänmaa

;; This file is part of Theme-D-Intr.

(define-module (guile-theme-d-intr support3))

(export theme-d-xml->sxml)

(import (theme-d runtime theme-d-stdlib-support)
        (sxml simple))

(define (theme-d-xml->sxml str-xml)
  (let ((sx (xml->sxml str-xml #:trim-whitespace? #t)))
    (check-source-expr sx)
    sx))
